/*
	Import Diablo III Animation
	Written by Taylor Mouse (c) 01.2014
	
*/
ClearListener()

/* FUNCTIONS */
fn ReadBytes stream val =
(	
	for i=1 to val do ReadByte stream
)

fn ReadFixedString stream val =
(
	local str = ""
	for i=1 to val do
	(
		s= bit.IntAsChar(ReadByte stream)
		
		if(s!="\0") then str+=s
	)
	return str
)

fn SetWorldRotation theNode theRot = 
( 
	local matrix = transmatrix theNode.transform.pos
	in coordsys matrix (theNode.rotation = theRot)
) 

/* STRUCTS */
struct D3Bone ( ID, Name, TheBone )
struct TranslationAnimation ( BoneId, KeyFrame, Position )
struct RotationAnimation (BoneId, KeyFrame, Quaternion  )
struct ScaleAnimation (BoneId, KeyFrame, Scaling )

file = getOpenFileName \ 
	caption:"Diablo 3 Animation File" \
	types:"Diablo 3 Animation File(*.ani)|*.ani" \
	historyCategory:"Diablo3AnimationPresets"

if (file == undefined) then return undefined

Format "-- START READING DIABLO III ANIMATION FILE(s) --"

stream = fOpen file "rb"	-- Open the file for reading

ReadBytes stream 76 		-- Skip 76 bytes

animationName = ReadFixedString stream 64	-- Animation name

ReadBytes stream 52 		-- Skip 52 bytes

nBones = ReadLong stream	-- number of bones
offBones = ReadLong stream	-- data for the bones
sizeBones = ReadLong stream -- size in bytes of the bones 

ReadBytes stream 12 		-- Skip 12 bytes

totalFrames = ReadLong stream	-- total frames  in this animation

offTranslation = ReadLong stream 	-- data for translation animation
sizeTranslation = ReadLong stream	-- size of the translation animation data

ReadBytes stream 12 		-- Skip 12 bytes

offRotation = ReadLong stream 		-- data for the rotations 
sizeRotation = ReadLong stream 		-- size of the rotation data

ReadBytes stream 8 			-- Skip 8 bytes

offScale = ReadLong stream 		-- data for the scaling
sizeScale = ReadLong stream 	-- size for the scaling data

ReadBytes stream 208 			-- Skip 208 bytes

xan = ReadFixedString stream 256	-- always 0000.xan
source = ReadFixedString stream 256	-- initial location of this file

/* BONES */
fSeek stream ( offBones + 16 ) #seek_set -- jump to the bone data position
allBones = #()
for i=1 to nBones do
(
	b = D3Bone()
	b.ID = i	--> for reference only
	b.Name = ReadFixedString stream 64	-- name of the bone
	b.TheBone = GetNodeByName b.Name	-- find the bone in the scene
	append allBones b
)
print allBones

/* Frame data*/
nFrameData = #()
offFrameData = #()
sizeFrameData = #()

/* TRANSLATION */
fSeek stream ( offTranslation + 16 ) #seek_set 
for i=1 to nBones do -- get the offset of where the translation data is stored
(
	append nFrameData (ReadLong stream)
	append offFrameData (ReadLong stream)
	append sizeFrameData (ReadLong stream)
	ReadBytes stream 12	-- always 00 00 00 00 - 00 00 00 00 - 00 00 00 00 ?
)

allTranslations = #()
for i=1 to nBones do -- Read the translation data structure
(	
	fSeek stream ( offFrameData[i] + 16 ) #seek_set 
	
	for fd=1 to nFrameData[i] do
	(
		tran = TranslationAnimation()
		tran.BoneId = i
		tran.KeyFrame = ReadLong stream
		tran.Position = [ReadFloat stream, ReadFloat stream, ReadFloat stream]

		append allTranslations tran
	)
)
--print allTranslations

/* Reset Frame data*/
nFrameData = #()
offFrameData = #()
sizeFrameData = #()

/* ROTATION */
fSeek stream ( offRotation + 16 ) #seek_set 
for i=1 to nBones do -- get the offset of where the translation data is stored
(
	append nFrameData (ReadLong stream)
	append offFrameData (ReadLong stream)
	append sizeFrameData (ReadLong stream)
	ReadBytes stream 12	-- always 00 00 00 00 - 00 00 00 00 - 00 00 00 00 ?
)

allRotations = #()
for i=1 to nBones do -- Read the rotation data structure
(	
	fSeek stream ( offFrameData[i] + 16 ) #seek_set 
	
	for fd=1 to nFrameData[i] do
	(
		rot = RotationAnimation()
		rot.BoneId = i
		rot.KeyFrame = ReadLong stream
		x = (ReadShort stream #signed) / 32767.0
		y = (ReadShort stream #signed) / 32767.0
		z = (ReadShort stream #signed) / 32767.0
		w = (ReadShort stream #signed) / 32767.0
		
		rot.Quaternion = quat x y z w
	
		append allRotations rot
	)
)
--print allRotations

/* Reset Frame data*/
nFrameData = #()
offFrameData = #()
sizeFrameData = #()

/* SCALE */
fSeek stream ( offScale + 16 ) #seek_set 
for i=1 to nBones do -- get the offset of where the scaling data is stored
(
	append nFrameData (ReadLong stream)
	append offFrameData (ReadLong stream)
	append sizeFrameData (ReadLong stream)
	ReadBytes stream 12	
)

allScaling = #()
for i=1 to nBones do -- Read the scale data structure
(	
	fSeek stream ( offFrameData[i] + 16 ) #seek_set 
	
	for fd=1 to nFrameData[i] do
	(
		s = ScaleAnimation()
		s.BoneId = i
		s.KeyFrame = ReadLong stream
		s.Scaling = ReadFloat stream

		append allScaling s
	)
)
--print allScaling
echo "-- DONE READING ANI FILE --\n"
echo "-- APPLYING ANIMATION TO THE BONES --\n"

-- TODO:  get the last used animation key, so we can add multiple animations to the scene
startingFrame = 10

-- Apply the Translation
for i=1 to allTranslations.count do
(
	currentBone = allBones[allTranslations[i].BoneId].TheBone
	parentBone = undefined
	parentBoneId = 0
	parentPosition = [0,0,0]
	if(currentBone == undefined ) then continue
	
	-- get the parent bone position in this frame, because the bone is reference to its parent position
	if( currentBone.Parent != undefined ) then 
	(
		parentBone = currentBone.Parent
		for x=1 to nBones do
		(
			if( allBones[x].TheBone.Name == parentBone.Name ) then
			(
				parentPosition = allBones[x].TheBone.Position
				exit -- the for loop
			)
			else 
			(
				parentPosition = [0,0,0]
			)
		)
	)
	
	keyFrame = allTranslations[i].KeyFrame + startingFrame
	posAnim = allTranslations[i].Position + parentPosition

	at time keyFrame
		animate on 
			currentBone.pos = posAnim

)

--Apply the Rotation
prevQ = quat 1 -- set a starting rotation
prevB = 1
for i=1 to allRotations.count do
(
	b = allBones[allRotations[i].BoneId].TheBone
	if(b == undefined) then continue
	
	if(prevB!=b) then
	(
		prevQ = quat 1
		prevB = b
	)
	
	t = allRotations[i].KeyFrame + startingFrame 
	a = QuatToEuler(allRotations[i].Quaternion - prevQ)
	a.X *=-1
	a.Y *=-1
	a.Z *=-1
	prevQ = allRotations[i].Quaternion
	
	set time t
		set animate on
			rotate b a
		
)

-- Apply the Scaling
fClose stream

